/**
 * @file   config.h
 *
 * @section LICENSE
 *
 * The MIT License
 *
 * @copyright Copyright (c) 2017-2019 TileDB, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * @section DESCRIPTION
 *
 * This file defines class Config.
 */

#ifndef TILEDB_CONFIG_H
#define TILEDB_CONFIG_H

#include "tiledb/sm/misc/constants.h"
#include "tiledb/sm/misc/status.h"

#include <map>
#include <set>
#include <thread>

namespace tiledb {
namespace sm {

/** This class manages the TileDB configuration options. */
class Config {
 public:
  /* ********************************* */
  /*         TYPE DEFINITIONS          */
  /* ********************************* */

  /** Consolidation parameters. */
  struct ConsolidationParams {
    float amplification_;
    uint64_t buffer_size_;
    uint32_t steps_;
    uint32_t step_min_frags_;
    uint32_t step_max_frags_;
    float step_size_ratio_;

    ConsolidationParams() {
      amplification_ = constants::consolidation_amplification;
      steps_ = constants::consolidation_steps;
      buffer_size_ = constants::consolidation_buffer_size;
      step_min_frags_ = constants::consolidation_step_min_frags;
      step_max_frags_ = constants::consolidation_step_max_frags;
      step_size_ratio_ = constants::consolidation_step_size_ratio;
    }
  };

  struct RESTParams {
    std::string server_address_;
    SerializationType server_serialization_format_;
    std::string username_;
    std::string password_;
    std::string token_;

    RESTParams() {
      server_address_ = "";
      server_serialization_format_ = constants::serialization_default_format;
      username_ = "";
      password_ = "";
      token_ = "";
    }
  };

  /** Storage manager parameters. */
  struct SMParams {
    bool enable_signal_handlers_;
    uint64_t memory_budget_;
    uint64_t memory_budget_var_;
    uint64_t num_async_threads_;
    uint64_t num_reader_threads_;
    uint64_t num_writer_threads_;
    int num_tbb_threads_;
    uint64_t tile_cache_size_;
    bool dedup_coords_;
    bool check_coord_dups_;
    bool check_coord_oob_;
    bool check_global_order_;
    ConsolidationParams consolidation_params_;

    SMParams() {
      enable_signal_handlers_ = constants::enable_signal_handlers;
      memory_budget_ = constants::memory_budget_fixed;
      memory_budget_var_ = constants::memory_budget_var;
      num_async_threads_ = constants::num_async_threads;
      num_reader_threads_ = constants::num_reader_threads;
      num_writer_threads_ = constants::num_writer_threads;
      num_tbb_threads_ = constants::num_tbb_threads;
      tile_cache_size_ = constants::tile_cache_size;
      dedup_coords_ = false;
      check_coord_dups_ = true;
      check_coord_oob_ = true;
      check_global_order_ = true;
    }
  };

  struct S3Params {
    std::string region_;
    std::string scheme_;
    std::string endpoint_override_;
    bool use_virtual_addressing_;
    bool use_multipart_upload_;
    uint64_t max_parallel_ops_;
    uint64_t multipart_part_size_;
    long connect_timeout_ms_;
    long connect_max_tries_;
    long connect_scale_factor_;
    long request_timeout_ms_;
    std::string proxy_scheme_;
    std::string proxy_host_;
    unsigned proxy_port_;
    std::string proxy_username_;
    std::string proxy_password_;
    std::string aws_access_key_id;
    std::string aws_secret_access_key;
    std::string logging_level_;

    S3Params() {
      region_ = constants::s3_region;
      scheme_ = constants::s3_scheme;
      endpoint_override_ = constants::s3_endpoint_override;
      use_virtual_addressing_ = constants::s3_use_virtual_addressing;
      use_multipart_upload_ = constants::s3_use_multipart_upload;
      max_parallel_ops_ = constants::s3_max_parallel_ops;
      multipart_part_size_ = constants::s3_multipart_part_size;
      connect_timeout_ms_ = constants::s3_connect_timeout_ms;
      connect_max_tries_ = constants::s3_connect_max_tries;
      connect_scale_factor_ = constants::s3_connect_scale_factor;
      request_timeout_ms_ = constants::s3_request_timeout_ms;
      proxy_scheme_ = constants::s3_proxy_scheme;
      proxy_host_ = constants::s3_proxy_host;
      proxy_port_ = constants::s3_proxy_port;
      proxy_username_ = constants::s3_proxy_username;
      proxy_password_ = constants::s3_proxy_password;
      aws_access_key_id = constants::aws_access_key_id;
      aws_secret_access_key = constants::aws_secret_access_key;
      logging_level_ = constants::s3_logging_level;
    }
  };

  struct HDFSParams {
    std::string name_node_uri_;
    std::string username_;
    std::string kerb_ticket_cache_path_;

    HDFSParams() {
      kerb_ticket_cache_path_ = constants::hdfs_kerb_ticket_cache_path;
      name_node_uri_ = constants::hdfs_name_node_uri;
      username_ = constants::hdfs_username;
    }
  };

  struct FileParams {
    uint64_t max_parallel_ops_;
    bool enable_filelocks_;

    FileParams() {
      max_parallel_ops_ = constants::vfs_file_max_parallel_ops;
      enable_filelocks_ = constants::vfs_file_enable_filelocks;
    }
  };

  struct VFSParams {
    S3Params s3_params_;
    HDFSParams hdfs_params_;
    FileParams file_params_;
    uint64_t num_threads_;
    uint64_t min_parallel_size_;
    uint64_t min_batch_gap_;
    uint64_t min_batch_size_;

    VFSParams() {
      num_threads_ = constants::vfs_num_threads;
      min_batch_gap_ = constants::vfs_min_batch_gap;
      min_batch_size_ = constants::vfs_min_batch_size;
      min_parallel_size_ = constants::vfs_min_parallel_size;
    }
  };

  /* ********************************* */
  /*     CONSTRUCTORS & DESTRUCTORS    */
  /* ********************************* */

  /** Constructor. */
  Config();

  /** Constructor. */
  Config(const VFSParams& vfs_params);

  /** Destructor. */
  ~Config();

  /* ********************************* */
  /*                API                */
  /* ********************************* */

  /** Loads the config parameters from a configuration (local) file. */
  Status load_from_file(const std::string& filename);

  /** Saves the config parameters to a configuration (local) file. */
  Status save_to_file(const std::string& filename);

  /** Returns the rest interface parameters. */
  RESTParams rest_params() const;

  /** Returns the storage manager parameters. */
  SMParams sm_params() const;

  /** Returns the consolidation parameters. */
  ConsolidationParams consolidation_params() const;

  /** Returns the VFS parameters. */
  VFSParams vfs_params() const;

  /** Returns the S3 parameters. */
  S3Params s3_params() const;

  /** Sets a config parameter.
   *
   * **Parameters**
   *
   * - `sm.dedup_coords` <br>
   *    If `true`, cells with duplicate coordinates will be removed during
   *    sparse array writes. Note that ties during deduplication are
   *    arbitrary. <br>
   *    **Default**: false
   * - `sm.check_coord_dups` <br>
   *    This is applicable only if `sm.dedup_coords` is `false`.
   *    If `true`, an error will be thrown if there are cells with duplicate
   *    coordinates during sparse array writes. If `false` and there are
   *    duplicates, the duplicates will be written without errors, but the
   *    TileDB behavior could be unpredictable. <br>
   *    **Default**: true
   * - `sm.check_coord_oob` <br>
   *    If `true`, an error will be thrown if there are cells with coordinates
   *    falling outside the array domain during sparse array writes. <br>
   *    **Default**: true
   * - `sm.check_global_order` <br>
   *    Checks if the coordinates obey the global array order. Applicable only
   *    to sparse writes in global order.
   *    **Default**: true
   * - `sm.tile_cache_size` <br>
   *    The tile cache size in bytes. Any `uint64_t` value is acceptable. <br>
   *    **Default**: 10,000,000
   * - `sm.enable_signal_handlers` <br>
   *    Whether or not TileDB will install signal handlers. <br>
   *    **Default**: true
   * - `sm.num_async_threads` <br>
   *    The number of threads allocated for async queries. <br>
   *    **Default**: 1
   * - `sm.num_reader_threads` <br>
   *    The number of threads allocated for issuing reads to VFS in
   *    parallel. <br>
   *    **Default**: 1
   * - `sm.num_writer_threads` <br>
   *    The number of threads allocated for issuing writes to VFS in
   *    parallel.<br>
   *    **Default**: 1
   * - `sm.num_tbb_threads` <br>
   *    The number of threads allocated for the TBB thread pool (if TBB is
   *    enabled). Note: this is a whole-program setting. Usually this should not
   *    be modified from the default. See also the documentation for TBB's
   *    `task_scheduler_init` class.<br>
   *    **Default**: TBB automatic
   * - `sm.consolidation.amplification` <br>
   *    The factor by which the size of the dense fragment resulting
   *    from consolidating a set of fragments (containing at least one
   *    dense fragment) can be amplified. This is important when
   *    the union of the non-empty domains of the fragments to be
   *    consolidated have a lot of empty cells, which the consolidated
   *    fragment will have to fill with the special fill value
   *    (since the resulting fragments is dense). <br>
   *    **Default**: FLOAT_MAX
   * - `sm.consolidation.buffer_size` <br>
   *    The size (in bytes) of the attribute buffers used during
   *    consolidation. <br>
   *    **Default**: 50,000,000
   * - `sm.consolidation.steps` <br>
   *    The number of consolidation steps to be performed when executing
   *    the consolidation algorithm.<br>
   *    **Default**: 1
   * - `sm.consolidation.step_min_frags` <br>
   *    The minimum number of fragments to consolidate in a single step.<br>
   *    **Default**: UINT32_MAX
   * - `sm.consolidation.step_max_frags` <br>
   *    The maximum number of fragments to consolidate in a single step.<br>
   *    **Default**: UINT32_MAX
   * - `sm.consolidation.step_size_ratio` <br>
   *    The size ratio that two ("adjacent") fragments must satisfy to be
   *    considered for consolidation in a single step.<br>
   *    **Default**: 0.0
   * - `sm.memory_budget` <br>
   *    The memory budget for tiles of fixed-sized attributes (or offsets for
   *    var-sized attributes) to be fetched during reads.<br>
   *    **Default**: 5GB
   * - `sm.memory_budget_var` <br>
   *    The memory budget for tiles of var-sized attributes
   *    to be fetched during reads.<br>
   *    **Default**: 10GB
   * - `vfs.num_threads` <br>
   *    The number of threads allocated for VFS operations (any backend), per
   *    VFS instance. <br>
   *    **Default**: number of cores
   * - `vfs.min_parallel_size` <br>
   *    The minimum number of bytes in a parallel VFS operation
   *    (except parallel S3 writes, which are controlled by
   *    `vfs.s3.multipart_part_size`.) <br>
   *    **Default**: 10MB
   * - `vfs.min_batch_gap` <br>
   *    The minimum number of bytes between two VFS read batches.<br>
   *    **Default**: 500KB
   * - `vfs.min_batch_size` <br>
   *    The minimum number of bytes in a VFS read operation<br>
   *    **Default**: 20MB
   * - `vfs.file.max_parallel_ops` <br>
   *    The maximum number of parallel operations on objects with `file:///`
   *    URIs. <br>
   *    **Default**: `vfs.num_threads`
   * - `vfs.file.enable_filelocks` <br>
   *    If set to `false`, file locking operations are no-ops for `file:///`
   *    URIs in VFS. <br>
   *    **Default**: `true`
   * - `vfs.s3.region` <br>
   *    The S3 region, if S3 is enabled. <br>
   *    **Default**: us-east-1
   * - `vfs.s3.aws_access_key_id` <br>
   *    Set the AWS_ACCESS_KEY_ID <br>
   *    **Default**: ""
   * - `vfs.s3.aws_secret_access_key` <br>
   *    Set the AWS_SECRET_ACCESS_KEY <br>
   *    **Default**: ""
   * - `vfs.s3.scheme` <br>
   *    The S3 scheme (`http` or `https`), if S3 is enabled. <br>
   *    **Default**: https
   * - `vfs.s3.endpoint_override` <br>
   *    The S3 endpoint, if S3 is enabled. <br>
   *    **Default**: ""
   * - `vfs.s3.use_virtual_addressing` <br>
   *    The S3 use of virtual addressing (`true` or `false`), if S3 is
   *    enabled. <br>
   *    **Default**: true
   * - `vfs.s3.use_multipart_upload` <br>
   *    The S3 use of multipart upload (`true` or `false`), if S3 is
   *    enabled. <br>
   *    **Default**: true
   * - `vfs.s3.max_parallel_ops` <br>
   *    The maximum number of S3 backend parallel operations. <br>
   *    **Default**: `vfs.num_threads`
   * - `vfs.s3.multipart_part_size` <br>
   *    The part size (in bytes) used in S3 multipart writes.
   *    Any `uint64_t` value is acceptable. Note: `vfs.s3.multipart_part_size *
   *    vfs.s3.max_parallel_ops` bytes will be buffered before issuing multipart
   *    uploads in parallel. <br>
   *    **Default**: 5MB
   * - `vfs.s3.connect_timeout_ms` <br>
   *    The connection timeout in ms. Any `long` value is acceptable. <br>
   *    **Default**: 3000
   * - `vfs.s3.connect_max_tries` <br>
   *    The maximum tries for a connection. Any `long` value is acceptable. <br>
   *    **Default**: 5
   * - `vfs.s3.connect_scale_factor` <br>
   *    The scale factor for exponential backofff when connecting to S3.
   *    Any `long` value is acceptable. <br>
   *    **Default**: 25
   * - `vfs.s3.logging_level` <br>
   *    The AWS SDK logging level. This is a process-global setting. The
   *    configuration of the most recently constructed context will set
   *    process state. Log files are written to the process working directory.
   *    **Default**: ""
   * - `vfs.s3.request_timeout_ms` <br>
   *    The request timeout in ms. Any `long` value is acceptable. <br>
   *    **Default**: 3000
   * - `vfs.s3.proxy_host` <br>
   *    The proxy host. <br>
   *    **Default**: ""
   * - `vfs.s3.proxy_port` <br>
   *    The proxy port. <br>
   *    **Default**: 0
   * - `vfs.s3.proxy_scheme` <br>
   *    The proxy scheme. <br>
   *    **Default**: "https"
   * - `vfs.s3.proxy_username` <br>
   *    The proxy username. Note: this parameter is not serialized by
   *    `tiledb_config_save_to_file`. <br>
   *    **Default**: ""
   * - `vfs.s3.proxy_password` <br>
   *    The proxy password. Note: this parameter is not serialized by
   *    `tiledb_config_save_to_file`. <br>
   *    **Default**: ""
   * - `vfs.hdfs.name_node"` <br>
   *    Name node for HDFS. <br>
   *    **Default**: ""
   * - `vfs.hdfs.username` <br>
   *    HDFS username. <br>
   *    **Default**: ""
   * - `vfs.hdfs.kerb_ticket_cache_path` <br>
   *    HDFS kerb ticket cache path. <br>
   *    **Default**: ""
   *
   * <br>
   *
   * - `rest.server_address` <br>
   *    URL for REST server to use for remote arrays. <br>
   *    **Default**: ""
   * - `rest.server_serialization_format` <br>
   *    Serialization format to use for remote array requests (CAPNP or
   *    JSON). <br>
   *    **Default**: "CAPNP"
   * - `rest.username` <br>
   *    Username for login to REST server (a token can be used instead). <br>
   *    **Default**: ""
   * - `rest.password` <br>
   *    Password for login to REST server. <br>
   *    **Default**: ""
   * - `rest.token` <br>
   *    Authentication token for REST server (used instead of
   *    username/password). <br>
   *    **Default**: ""
   */
  Status set(const std::string& param, const std::string& value);

  /** Gets a config parameter value (`nullptr` if `param` does not exist). */
  Status get(const std::string& param, const char** value) const;

  /**
   * Returns a map with the param-value pairs in the config.
   *
   * @param prefix If equal to an empty string, all param-value pairs are
   *     retrieved. Otherwise, only those that start with `prefix.*` will
   *     be retrieved noting that the prefix will be stripped from the
   *     returned params.
   * @return A map with the param-value.
   */
  std::map<std::string, std::string> param_values(
      const std::string& prefix) const;

  /** Unsets a parameter (ignores it if it does not exist). */
  Status unset(const std::string& param);

 private:
  /* ********************************* */
  /*         PRIVATE ATTRIBUTES        */
  /* ********************************* */

  /** Stores a map of param -> value. */
  std::map<std::string, std::string> param_values_;

  /** The rest interface parameters. */
  RESTParams rest_params_;

  /** The storage manager parameters. */
  SMParams sm_params_;

  /** The VFS parameters. */
  VFSParams vfs_params_;

  /* ********************************* */
  /*          PRIVATE CONSTANTS        */
  /* ********************************* */

  /** Character indicating the start of a comment in a config file. */
  static const char COMMENT_START;

  /** Set of parameter names that are not serialized to file. */
  static const std::set<std::string> unserialized_params_;

  /* ********************************* */
  /*          PRIVATE METHODS          */
  /* ********************************* */

  /** Populates `param_values_` map with the default values. */
  void set_default_param_values();

  /** Sets the dedup coordinates parameter. */
  Status set_sm_dedup_coords(const std::string& value);

  /** Sets the check for coordinates duplicates parameter. */
  Status set_sm_check_coord_dups(const std::string& value);

  /** Sets the check for out-of-bound coordinates parameter. */
  Status set_sm_check_coord_oob(const std::string& value);

  /** Sets the check for global order parameter. */
  Status set_sm_check_global_order(const std::string& value);

  /** Sets the enable signal handlers value, properly parsing the input value.*/
  Status set_sm_enable_signal_handlers(const std::string& value);

  /** Sets the number of threads, properly parsing the input value.*/
  Status set_sm_num_async_threads(const std::string& value);

  /** Sets the number of threads, properly parsing the input value.*/
  Status set_sm_num_reader_threads(const std::string& value);

  /** Sets the number of threads, properly parsing the input value.*/
  Status set_sm_num_writer_threads(const std::string& value);

  /** Sets the number of TBB threads, properly parsing the input value.*/
  Status set_sm_num_tbb_threads(const std::string& value);

  /** Sets the number of consolidation steps.*/
  Status set_consolidation_steps(const std::string& value);

  /** Sets the minimum number of fragments to consolidate in a step.*/
  Status set_consolidation_step_min_frags(const std::string& value);

  /** Sets the maximum number of fragments to consolidate in a step.*/
  Status set_consolidation_step_max_frags(const std::string& value);

  /** Sets the size ratio requirement for two "adjacent" fragments in a step. */
  Status set_consolidation_step_size_ratio(const std::string& value);

  /** Sets the consolidation amplification, properly parsing the input value. */
  Status set_consolidation_amplification(const std::string& value);

  /** Sets the consolidation buffer size, properly parsing the input value. */
  Status set_consolidation_buffer_size(const std::string& value);

  /** Sets the memory_budget, properly parsing the input value. */
  Status set_sm_memory_budget(const std::string& value);

  /**
   * Sets the memory_budget for var-sized attributes, properly parsing the
   * input value.
   */
  Status set_sm_memory_budget_var(const std::string& value);

  /** Sets the tile cache size, properly parsing the input value. */
  Status set_sm_tile_cache_size(const std::string& value);

  /** Set the rest server address */
  Status set_rest_server_address(const std::string& value);

  /** Set the rest server serialization format */
  Status set_rest_server_serialization_format(const std::string& value);

  /** Set the rest server username for auth */
  Status set_rest_username(const std::string& value);

  /** Set the rest server password for auth */
  Status set_rest_password(const std::string& value);

  /** Set the rest server token for auth */
  Status set_rest_token(const std::string& value);

  /** Sets the number of VFS threads. */
  Status set_vfs_num_threads(const std::string& value);

  /** Sets the min number of bytes of a VFS parallel operation. */
  Status set_vfs_min_parallel_size(const std::string& value);

  /** Sets the min number of bytes between two VFS read batches. */
  Status set_vfs_min_batch_gap(const std::string& value);

  /** Sets the min number of bytes of a VFS read operation. */
  Status set_vfs_min_batch_size(const std::string& value);

  /** Sets the max number of allowed file:/// parallel operations. */
  Status set_vfs_file_max_parallel_ops(const std::string& value);

  /** Sets the file locking enabled state. */
  Status set_vfs_file_enable_filelocks(const std::string& value);

  /** Sets the S3 region. */
  Status set_vfs_s3_region(const std::string& value);

  /** Sets the S3 AWS_ACCESS_KEY_ID. */
  Status set_vfs_s3_aws_access_key_id(const std::string& value);

  /** Sets the S3 AWS_SECRET_ACCESS_KEY. */
  Status set_vfs_s3_aws_secret_access_key(const std::string& value);

  /** Sets the S3 scheme. */
  Status set_vfs_s3_scheme(const std::string& value);

  /** Sets the S3 endpoint override. */
  Status set_vfs_s3_endpoint_override(const std::string& value);

  /** Sets the S3 virtual addressing. */
  Status set_vfs_s3_use_virtual_addressing(const std::string& value);

  /** Sets the S3 multipart upload boolean. */
  Status set_vfs_s3_use_multipart_upload(const std::string& value);

  /** Sets the maximum number of parallel S3 operations. */
  Status set_vfs_s3_max_parallel_ops(const std::string& value);

  /** Sets the S3 multipart part size. */
  Status set_vfs_s3_multipart_part_size(const std::string& value);

  /** Sets the S3 connect timeout in milliseconds. */
  Status set_vfs_s3_connect_timeout_ms(const std::string& value);

  /** Sets the S3 connect maximum tries. */
  Status set_vfs_s3_connect_max_tries(const std::string& value);

  /** Sets the S3 connect scale factor for exponential backoff. */
  Status set_vfs_s3_connect_scale_factor(const std::string& value);

  /** Sets the S3 logging level. */
  Status set_vfs_s3_logging_level(const std::string& value);

  /** Sets the S3 request timeout in milliseconds. */
  Status set_vfs_s3_request_timeout_ms(const std::string& value);

  /** Sets the S3 proxy scheme. */
  Status set_vfs_s3_proxy_scheme(const std::string& value);

  /** Sets the S3 proxy host. */
  Status set_vfs_s3_proxy_host(const std::string& value);

  /** Sets the S3 proxy port. */
  Status set_vfs_s3_proxy_port(const std::string& value);

  /** Sets the S3 proxy username. */
  Status set_vfs_s3_proxy_username(const std::string& value);

  /** Sets the S3 proxy password. */
  Status set_vfs_s3_proxy_password(const std::string& value);

  /** Sets the HDFS namenode hostname and port (uri) */
  Status set_vfs_hdfs_name_node(const std::string& value);

  /** Sets the HDFS username */
  Status set_vfs_hdfs_username(const std::string& value);

  /** Sets the Kerberos auth ticket cache path */
  Status set_vfs_hdfs_kerb_ticket_cache_path(const std::string& value);
};

}  // namespace sm
}  // namespace tiledb

#endif  // TILEDB_CONFIG_H
